/*
 * TouchGraph LLC. Apache-Style Software License
 *
 *
 * Copyright (c) 2001-2002 Alexander Shapiro. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:  
 *       "This product includes software developed by 
 *        TouchGraph LLC (http://www.touchgraph.com/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "TouchGraph" or "TouchGraph LLC" must not be used to endorse 
 *    or promote products derived from this software without prior written 
 *    permission.  For written permission, please contact 
 *    alex@touchgraph.com
 *
 * 5. Products derived from this software may not be called "TouchGraph",
 *    nor may "TouchGraph" appear in their name, without prior written
 *    permission of alex@touchgraph.com.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL TOUCHGRAPH OR ITS CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR 
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, 
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 */

package com.touchgraph.graphlayout;

import  com.touchgraph.graphlayout.interaction.*;

import  java.awt.*;
import  java.awt.event.*;
import  javax.swing.*;
import  java.util.Hashtable;

/** GLPanel contains code for adding scrollbars and interfaces to the TGPanel
  * The "GL" prefix indicates that this class is GraphLayout specific, and
  * will probably need to be rewritten for other applications.
  *
  * @author   Alexander Shapiro
  * @version  1.21  $Id: GLPanel.java,v 1.31 2002/04/04 06:19:49 x_ander Exp $
  */
public class GLPanel extends JPanel {

    public String zoomLabel = "Zoom"; // label for zoom menu item
    public String rotateLabel = "Rotate"; // label for rotate menu item
    public String localityLabel = "Locality"; // label for locality menu item

    public HVScroll hvScroll;
    public ZoomScroll zoomScroll;
    //public HyperScroll hyperScroll; // unused
    public RotateScroll rotateScroll;
    public LocalityScroll localityScroll;
    public JPopupMenu glPopup;
    public Hashtable scrollBarHash; //= new Hashtable();

    protected TGPanel tgPanel;
    protected TGLensSet tgLensSet;
    protected TGUIManager tgUIManager;

    private Color defaultColor = Color.lightGray;

    private String [] labels;
    private int [] ass;
    private double [] strength;
    private double [][] dist;
    private int K;
    private int N;


  // ............


   /** Default constructor.
     */
    public GLPanel() {
        scrollBarHash = new Hashtable();
        tgLensSet = new TGLensSet();
        tgPanel = new TGPanel();
        hvScroll = new HVScroll(tgPanel, tgLensSet);
        zoomScroll = new ZoomScroll(tgPanel);
      //hyperScroll = new HyperScroll(tgPanel);
        rotateScroll = new RotateScroll(tgPanel);
        localityScroll = new LocalityScroll(tgPanel);
        initialize();
    }

    public GLPanel(Color color, String [] l, int [] a, double [] s, int NN, double [][] d, int KK) {
	defaultColor = color;
        this.setBackground(color);
        scrollBarHash = new Hashtable();
        tgLensSet = new TGLensSet();
        tgPanel = new TGPanel();
        hvScroll = new HVScroll(tgPanel, tgLensSet);
        zoomScroll = new ZoomScroll(tgPanel);
      //hyperScroll = new HyperScroll(tgPanel);
        rotateScroll = new RotateScroll(tgPanel);
        localityScroll = new LocalityScroll(tgPanel);
	labels = l;
	ass = a;
	strength = s;
	N = NN;
	dist = d;
	K = KK;
        initialize();
    }




   /** Constructor with a Color to be used for UI background.
     */
    public GLPanel( Color color ) {
        defaultColor = color;
        this.setBackground(color);
        scrollBarHash = new Hashtable();
        tgLensSet = new TGLensSet();
        tgPanel = new TGPanel();
        tgPanel.setBackground(color);
        hvScroll = new HVScroll(tgPanel, tgLensSet);
      //hvScroll.getHorizontalSB().setBackground(Color.orange);
      //hvScroll.getVerticalSB().setBackground(Color.cyan);
        zoomScroll = new ZoomScroll(tgPanel);
      //zoomScroll.getZoomSB().setBackground(Color.green);
      //hyperScroll = new HyperScroll(tgPanel);
        rotateScroll = new RotateScroll(tgPanel);
      //rotateScroll.getRotateSB().setBackground(Color.blue);
        localityScroll = new LocalityScroll(tgPanel);
      //localityScroll.getLocalitySB().setBackground(Color.red);
        initialize();
    }


   /** Initialize panel, lens, and establish a random graph as a demonstration.
     */
    public void initialize() {
        buildPanel();
        buildLens();
        tgPanel.setLensSet(tgLensSet);
        addUIs();
      //tgPanel.addNode();  //Add a starting node.
        try {
            //randomGraph();
	    myGraph();
        } catch ( TGException tge ) {
            System.err.println(tge.getMessage());
            tge.printStackTrace(System.err);
        }
        tgPanel.setSelect(tgPanel.getGES().getFirstNode()); //Select first node, so hiding works
        setVisible(true);
    }

    /** Return the TGPanel used with this GLPanel. */
    public TGPanel getTGPanel() {
        return tgPanel;
    }

  // navigation .................

    /** Return the HVScroll used with this GLPanel. */
    public HVScroll getHVScroll()
    {
        return hvScroll;
    }

    ///** Return the HyperScroll used with this GLPanel. */
    //public HyperScroll getHyperScroll()
    //{
    //    return hyperScroll;
    //}

    /** Sets the horizontal offset to p.x, and the vertical offset to p.y
      * given a Point <tt>p<tt>. 
      */
    public void setOffset( Point p ) {
        hvScroll.setOffset(p);
    };

    /** Return the horizontal and vertical offset position as a Point. */
    public Point getOffset() {
        return hvScroll.getOffset();
    };

  // rotation ...................

    /** Return the RotateScroll used with this GLPanel. */
    public RotateScroll getRotateScroll()
    {
        return rotateScroll;
    }

    /** Set the rotation angle of this GLPanel (allowable values between 0 to 359). */
     public void setRotationAngle( int angle ) {
        rotateScroll.setRotationAngle(angle);
    }

    /** Return the rotation angle of this GLPanel. */
    public int getRotationAngle() {
        return rotateScroll.getRotationAngle();
    }

  // locality ...................

    /** Return the LocalityScroll used with this GLPanel. */
    public LocalityScroll getLocalityScroll()
    {
        return localityScroll;
    }

    /** Set the locality radius of this TGScrollPane  
      * (allowable values between 0 to 4, or LocalityUtils.INFINITE_LOCALITY_RADIUS). 
      */
    public void setLocalityRadius( int radius ) {
        localityScroll.setLocalityRadius(radius);
    }

    /** Return the locality radius of this GLPanel. */
    public int getLocalityRadius() {
        return localityScroll.getLocalityRadius();
    }

  // zoom .......................

    /** Return the ZoomScroll used with this GLPanel. */
    public ZoomScroll getZoomScroll() 
    {
        return zoomScroll;
    }

    /** Set the zoom value of this GLPanel (allowable values between -100 to 100). */
    public void setZoomValue( int zoomValue ) {
        zoomScroll.setZoomValue(zoomValue);
    }

    /** Return the zoom value of this GLPanel. */
    public int getZoomValue() {
        return zoomScroll.getZoomValue();
    }

  // ....

    public JPopupMenu getGLPopup() 
    {
        return glPopup;
    }

    public void buildLens() {
        tgLensSet.addLens(hvScroll.getLens());
        tgLensSet.addLens(zoomScroll.getLens());
      //tgLensSet.addLens(hyperScroll.getLens());
        tgLensSet.addLens(rotateScroll.getLens());
        tgLensSet.addLens(tgPanel.getAdjustOriginLens());
    }

    public void buildPanel() {
        final JScrollBar horizontalSB = hvScroll.getHorizontalSB();
        final JScrollBar verticalSB = hvScroll.getVerticalSB();
        final JScrollBar zoomSB = zoomScroll.getZoomSB();
        final JScrollBar rotateSB = rotateScroll.getRotateSB();
        final JScrollBar localitySB = localityScroll.getLocalitySB();

        setLayout(new BorderLayout());

        JPanel scrollPanel = new JPanel();
        scrollPanel.setBackground(defaultColor);
        scrollPanel.setLayout(new GridBagLayout());
        GridBagConstraints c = new GridBagConstraints();


	/*   JPanel modeSelectPanel = new JPanel();
        modeSelectPanel.setBackground(defaultColor);
        modeSelectPanel.setLayout(new FlowLayout(FlowLayout.CENTER, 0,0));/

	/*        AbstractAction navigateAction = new AbstractAction("Navigate") {
            public void actionPerformed(ActionEvent e) {
                tgUIManager.activate("Navigate");
            }
        };

        AbstractAction editAction = new AbstractAction("Edit") {
            public void actionPerformed(ActionEvent e) {
                tgUIManager.activate("Edit");
            }
	    };*/

	//  JRadioButton rbNavigate = new JRadioButton(navigateAction);
	//        rbNavigate.setBackground(defaultColor);
	//        rbNavigate.setSelected(true);
	//        JRadioButton rbEdit = new JRadioButton(editAction);
	//        rbEdit.setBackground(defaultColor);
	//        ButtonGroup bg = new ButtonGroup();
	//        bg.add(rbNavigate);
	//        bg.add(rbEdit);

	//        modeSelectPanel.add(rbNavigate);
	//        modeSelectPanel.add(rbEdit);

        final JPanel topPanel = new JPanel();
        topPanel.setBackground(defaultColor);
        topPanel.setLayout(new GridBagLayout());
        c.gridy=0; c.fill=GridBagConstraints.HORIZONTAL;
        /*
        c.gridx=0;c.weightx=0;
        topPanel.add(new Label("Zoom",Label.RIGHT), c);
        c.gridx=1;c.weightx=0.5;
        topPanel.add(zoomSB,c);
        c.gridx=2;c.weightx=0;
        topPanel.add(new Label("Locality",Label.RIGHT), c);
        c.gridx=3;c.weightx=0.5;
        topPanel.add(localitySB,c);
        */
        c.gridx=0;c.weightx=0;c.insets = new Insets(0,10,0,10);
	//        topPanel.add(modeSelectPanel,c);
        c.insets=new Insets(0,0,0,0);
        c.gridx=1;c.weightx=1;

        scrollBarHash.put(zoomLabel, zoomSB);
        scrollBarHash.put(rotateLabel, rotateSB);
        scrollBarHash.put(localityLabel, localitySB);

        JPanel scrollselect = scrollSelectPanel(new String[] {zoomLabel, rotateLabel, localityLabel});
        scrollselect.setBackground(defaultColor);
        topPanel.add(scrollselect,c);

        add(topPanel, BorderLayout.NORTH);

        c.fill = GridBagConstraints.BOTH;
        c.gridwidth = 1;
        c.gridx = 0; c.gridy = 1; c.weightx = 1; c.weighty = 1;
        scrollPanel.add(tgPanel,c);

        c.gridx = 1; c.gridy = 1; c.weightx = 0; c.weighty = 0;
        scrollPanel.add(verticalSB,c);

        c.gridx = 0; c.gridy = 2;
        scrollPanel.add(horizontalSB,c);

        add(scrollPanel,BorderLayout.CENTER);

        glPopup = new JPopupMenu();
        glPopup.setBackground(defaultColor);

        JMenuItem menuItem = new JMenuItem("Toggle Controls");
        ActionListener toggleControlsAction = new ActionListener() {
                boolean controlsVisible = true;
                public void actionPerformed(ActionEvent e) {
                    controlsVisible = !controlsVisible;
                    horizontalSB.setVisible(controlsVisible);
                    verticalSB.setVisible(controlsVisible);
                    topPanel.setVisible(controlsVisible);
                }
            };
        menuItem.addActionListener(toggleControlsAction);
        glPopup.add(menuItem);
    }

    protected JPanel scrollSelectPanel(String[] scrollBarNames) {
        final JComboBox scrollCombo = new JComboBox(scrollBarNames);
        scrollCombo.setBackground(defaultColor);
        scrollCombo.setPreferredSize(new Dimension(80,20));
        scrollCombo.setSelectedIndex(0);
        final JScrollBar initialSB = (JScrollBar) scrollBarHash.get(scrollBarNames[0]);
        scrollCombo.addActionListener(new ActionListener() {
            JScrollBar currentSB = initialSB;
            public void actionPerformed(ActionEvent e) {
                JScrollBar selectedSB = (JScrollBar) scrollBarHash.get(
                        (String) scrollCombo.getSelectedItem());
                if (currentSB!=null) currentSB.setVisible(false);
                if (selectedSB!=null) selectedSB.setVisible(true);
                currentSB = selectedSB;
            }
        });

        final JPanel sbp = new JPanel(new GridBagLayout());
        sbp.setBackground(defaultColor);
        GridBagConstraints c = new GridBagConstraints();
        c.gridx = 0; c.gridy = 0; c.weightx= 0;
        sbp.add(scrollCombo,c);
        c.gridx = 1; c.gridy = 0; c.weightx = 1; c.insets=new Insets(0,10,0,17);
        c.fill=GridBagConstraints.HORIZONTAL;
        for (int i = 0;i<scrollBarNames.length;i++) {
            JScrollBar sb = (JScrollBar) scrollBarHash.get(scrollBarNames[i]);
              if(sb==null) continue;
              if(i!=0) sb.setVisible(false);
              //sb.setMinimumSize(new Dimension(200,17));
              sbp.add(sb,c);
        }
        return sbp;
    }

    public void addUIs() {
        tgUIManager = new TGUIManager();
        GLEditUI editUI = new GLEditUI(this);
        GLNavigateUI navigateUI = new GLNavigateUI(this);
        tgUIManager.addUI(editUI,"Edit");
        tgUIManager.addUI(navigateUI,"Navigate");
        tgUIManager.activate("Navigate");
    }

    public void randomGraph() throws TGException {
        Node n1= tgPanel.addNode();
        n1.setType(0);
        for ( int i=0; i<246; i++ ) {
            Node r = tgPanel.getGES().getRandomNode();
            Node n = tgPanel.addNode();
            n.setType(0);
            if (tgPanel.findEdge(r,n)==null) tgPanel.addEdge(r,n,Edge.DEFAULT_LENGTH);
            if (i%2==0) {
                r = tgPanel.getGES().getRandomNode();
                if (tgPanel.findEdge(r,n)==null) tgPanel.addEdge(r,n,Edge.DEFAULT_LENGTH);
            }
        }
        tgPanel.setLocale(n1,2);
    }     

    public void myGraph() throws TGException {

	//	System.err.println("Maximum = " + maxa);    
	Node [] centre = new Node[K];
	Point [] p = new Point[K];

	MDS mds = new MDS(dist ,K); 
	double [][] points = mds.getPoints();
	for (int i=0; i<K; i++) {
	    // System.err.println((int)(1000.0*points[i][0]) + " " + (int)(1000.0*points[i][1]));
	    p[i] = new Point((int)(2000.0*points[i][0]),(int)(2000.0*points[i][1]));
	}
	
	for (int i=0; i<K; i++) {
	    centre[i] = tgPanel.addNode(null, new String("Cluster " + i), Color.black, p[i], i, true);
	    centre[i].setType(2);
	}

        for ( int i=0; i<N; i++ ) {
            Node n = tgPanel.addNode(labels[i], p[ass[i]], ass[i], false);
            n.setType(1);
            tgPanel.addEdge(n,centre[ass[i]],(int)(10.0+1000.0*strength[i]));
        }

	for (int i=0; i<K; i++) {
	    for (int j=0; j<i; j++) {
		tgPanel.addEdge2(centre[i],centre[j],(int)(10.0+2000.0*dist[i][j]), Color.black);
	    }
	}

	//	tgPanel.setLocale(centre[0],1);
    }  

    public static void main(String[] args) {

        JFrame frame;
        frame = new JFrame("Graph Layout");
        GLPanel glPanel = new GLPanel();
        frame.addWindowListener(new WindowAdapter() {
            public void windowClosing(WindowEvent e) {System.exit(0);}
        });

        frame.getContentPane().add("Center", glPanel);
        frame.setSize(500,500);
        frame.setVisible(true);
    }

} // end com.touchgraph.graphlayout.GLPanel





class MDS {

	private double [][] points;			
	private int dimension = 2;			
	private int iterations = 10;			// number of iterations of optimisation
	private double learnrate = 0.05;		// learning rate used in optimisation
	
	                                                // for normalisation:
	private double reshift;                      	// shift to origin
	private double rescale;                         // scale to interval [0, 1]
	
	private int its = 0;                         	// iteration counter
	private double [][] p;	                        // points working matrix
	private int N;                                  // number of data elements


	/** Constructor starting from readily extracted distance data
	* @param distances the precomputed inter-document distance data
	* @param N the number of inter-document distances
	* @param conf the current parameter settings
	*/
	public MDS(double [][] distances, int N) {
	
		try {
		    this.N = N;
		    for (int i=0; i<N; i++) {
			for (int j=i; j<N; j++) {
			    distances[i][j] = distances[j][i];
			}
		    }
		
		    // find minimum distance		
		    reshift = distances[1][0];
		    for (int i=0; i<N; i++) {
			for (int j=0; j<N; j++) {
			    reshift = Math.min(reshift, distances[i][j]);
			}
		    }
			// shift to origin
			for (int i=0; i<N; i++) {
				for (int j=0; j<N; j++) {
					distances[i][j] += -reshift;
				}
			}

			// find maximum distance
			rescale = distances[1][0];
			for (int i=0; i<N; i++) {
				for (int j=0; j<N; j++) {
					rescale = Math.max(rescale, distances[i][j]);
				}
			}
			// normalize to interval [0, 1]
			for (int i=0; i<N; i++) {
				for (int j=0; j<N; j++) {
					distances[i][j] *= 1.0/rescale;
				}
			}

			// initialisation
			points =  new double[N][dimension];
			p = new double[N][dimension];
			for (int i=0; i<N; i++) {
			    for (int j=0; j<dimension; j++) {
				p[i][j] = Math.random()*0.01-0.005;
				
			    }
			}
			
		} catch (Exception e) {
			System.err.println("MDS: Failure during preparations");
			System.err.println(e.getMessage());
			return;
		}
		
		try {
			
			int [] pinning_order;
		
			// main loop
			while (its<iterations) {
			    
			    its++;				
   				
			    // randomly permute the objects to determine the order
			    // in which they are pinned for this iteration
			    pinning_order = randperm(N);
   				
			    for (int i=0; i<N; i++) {

				int m = pinning_order[i];
					
      				// having pinned an object, move all of the other on each dimension
	      			// according to the learning rule
      				for (int j=0; j<N; j++) {
					
				    if (m != j) {
					double dummy = norm(m, j);
					
					if (dummy != 0) {
					    for (int k=0; k<dimension; k++) {
						p[j][k]=p[j][k]-learnrate*(dummy-distances[j][m])/dummy*Math.abs(p[j][k]-p[m][k])*sign(p[j][k]-p[m][k]);
									
					    }
					    
					}
				    }
				}       	
			    }
	        	}
	    	} catch (Exception ex) {
		    System.err.println("MDS: Failure during main optimisation loop");
		    System.err.println(ex.getMessage());
		    return;
		}
		
		for (int i=0; i<N; i++) {
		    for (int j=0; j<dimension; j++) {
			points[i][j] = p[i][j];//*rescale+reshift;
			
		    }
		}
		double [] mean = new double[dimension];
		for (int j=0; j<dimension; j++) {
		    mean[j] = 0.0;
		}
		for (int i=0; i<N; i++) {
		    for (int j=0; j<dimension; j++) {
			mean[j] += points[i][j];
		    }
		}
		for (int j=0; j<dimension; j++) {
		    mean[j] /= (double)(N);
		}

		for (int i=0; i<N; i++) {
		    for (int j=0; j<dimension; j++) {
			points[i][j] -= mean[j];
			
		    }
		}
		


	}

	// signum function
	private int sign(double value) {
		if (value == 0) return 0;
		else if (value > 0) return 1;
		else return -1;
	}

	// compute the Euclidean distance between the vectors represented by column m and j
	private double norm(int m, int j) {
		double sum = 0;
		for (int i=0; i<dimension; i++) {
			sum += (p[m][i] - p[j][i])*(p[m][i] - p[j][i]);
		}
	
		return Math.sqrt(sum);
	}

	// create a random ermutation of integers 0 to N-1
	private int[] randperm(int n) {
		int [] perm = new int[n];
		int [] numbers = new int[n];
		for (int i=0; i<n; i++) {
			numbers[i] = i+1;
		}
		for (int k=0; k<n; k++) {
			while (true) {
				int guess = (int)Math.floor(Math.random()*n);
				if (numbers[guess] != 0) {
					perm[k] = numbers[guess]-1;
					numbers[guess] = 0;
					break;
				}
			}
		}
		return perm;
	}


        /** Return the two-dimensional coordinates for the documents
	* @return a two-dimensional array of coordinates
	*/
	public double [][] getPoints() {
		return points;
	}
}


